;RS232 LCD Controller for plain-text
;===================================

.include "../include/tn2313def.inc"

;Controller:
;  ATTiny2313 @ 8 MHz (internal R/C-oscillator)
;  Low  Fuse temp = 0xC4 (8MHz R/C, BOD enabled)
;  High Fuse temp = 0xDB (BOD 2.7V)

;Ports:
;  PB0-7 = LCD Data 0..7
;  PD0   = RS232 Rx
;  PD1   = RS232 Tx (debugging only)
;  PD3   = LCD Backlight (active high)
;  PD4   = LCD E2 (dual-controller LCDs only)
;  PD5   = LCD RS
;  PD6   = LCD E

.equ	LCD_Data = PORTB
.equ	LCD_Ctrl = PORTD
.equ	LCD_E = 6
.equ	LCD_E2 = 4
.equ	LCD_RS = 5
.equ	LCD_BACKLIGHT = 3
.equ	BAUD = 51 ;Baud Rate: 207=2400, 103=4800, 51=9600, 34=14400, 25=19200
.equ	LCDLINES = 4 ;total lines (up to 8, modify linetab for more than 4)
.equ	LCDCOLS = 16 ;chars per line (up to 80)
.equ	DualController = 0

;RS232 Protocol:
;  * normal text (ASCII 0x20 to 0xff): send directly (automatic line break)
;  * control chars: supported control chars are
;    - 0x0a Line Feed: move cursor one line down and return cursor to column 0
;    - 0x0c Form Feed: clear LCD and move cursor home

;Version History:
;  2008-02-12  Arne Rossius
;    * first version
;  2010-04-10  Arne Rossius
;    * second version with slowed down timing
;  2010-08-20  Arne Rossius
;    * BUGFIX: clear display with 2 controllers
;    * BUGFIX: automatic line breaking when sending Line Feed
;    * added backlight output (off while LCD empty)

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	wait = R19
.def	line = R20 ;current LCD line
.def	col = R21 ;current LCD col
.def	bufptr = R22
.def	timeout = R23 ;timeout for transmission

;===============================================================================

.dseg
	Buffer: .byte LCDCOLS
	Stack: .byte 16

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	
;===============================================================================

linetab: ;memory locations of LCD lines
.if ( DualController )
	.db 0x00, 0x40, 0x00, 0x40
.else
	.db 0x00, 0x40, LCDCOLS, 0x40+LCDCOLS
.endif
	
lcdtext: ;LCD start-up message
	.db " RS232 LCD Ctrl "
	.db "================"
	.db "(C) Arne Rossius"
	.db " www.ebps.de.vu "

;===============================================================================

reset:
	;init stack pointer
	ldi	temp, RAMEND
	out	SPL, temp

	;init ports
	ldi	temp, 0xFF
	out	DDRB, temp
	ldi	temp, 0x00
	out	PORTB, temp
	
	ldi	temp, 0x7E
	out	DDRD, temp
	ldi	temp, 0x01
	out	PORTD, temp
	
	;init UART
	ldi	temp, 1<<RXEN | 1<<TXEN
	out	UCSRB, temp
	ldi	temp, BAUD
	out	UBRRL, temp
	
	;init LCD and display init message
	ldi	line, 0
	rcall	lcd_init
	.if ( DualController )
	ldi	line, LCDLINES/2
	rcall	lcd_init
	.endif
	ldi	ZL, LOW(lcdtext*2)
	ldi	ZH, HIGH(lcdtext*2)
	ldi	line, 0
initmessage1:
	push	ZL
	push	ZH
	ldi	col, 0
	rcall	lcd_line
	pop	ZH
	pop	ZL
	ldi	count, LCDCOLS
initmessage2:
	lpm	temp, Z+
	rcall	lcd_char
	dec	count
	brne	initmessage2
	inc	line
	cpi	line, LCDLINES
	brlo	initmessage1
	
	;init registers
	ldi	line, 0
	ldi	col, 0
	ldi	bufptr, Buffer
	
;-------------------------------------------------------------------------------

loop:
	;get char from UART, decide what to do (control char or data)
	rcall	getc
	cpi	temp, 0x0c ;Form Feed: Clear display, move cursor home
	breq	ff
	cpi	line, LCDLINES ;don't accept anything else if display is full
	brsh	loop
	
	;control chars
	cpi	temp, 0x00 ;End of String: display received string and lock
	breq	terminator
	cpi	temp, 0x0a ;Line Feed: Move cursor to beginning of next line
	breq	lf
	cpi	temp, 0x20 ;Space: display received string
	breq	space
	
	;printable chars
	cpi	temp, 0x09
	brlo	char_data ;LCD user defined char
	cpi	temp, 0x21
	brsh	char_data ;ASCII printable char (not space)
	
	;unknown/unimplemented char -> ignore
	rjmp	loop
	
;--------------------

ff:
	;form feed: clear display and buffer
	cbi	LCD_Ctrl, LCD_BACKLIGHT
	rcall	lcd_clear
	ldi	line, 0
	ldi	col, 0
	ldi	bufptr, Buffer
	rjmp	loop

;--------------------

lf:
	;line feed: display string buffer, then move to next line
	mov	count, bufptr
	subi	count, Buffer
	rcall	display_string
	inc	line
	ldi	col, 0
	rcall	lcd_line
	rjmp	loop
	
;--------------------

terminator:
	;terminator: display string buffer and lock display
	cpi	bufptr, Buffer
	breq	terminator_end
	mov	count, bufptr
	subi	count, Buffer
	rcall	display_string
terminator_end:
	ldi	line, LCDLINES ;lock display
	rjmp	loop

;--------------------

space:
	;space: display string buffer and a space
	cpi	bufptr, Buffer
	brne	space_string
	cpi	col, 0
	brne	space_char
	rjmp	loop ;don't display leading spaces
	
space_string:
	mov	count, bufptr
	subi	count, Buffer
	rcall	display_string
space_char:
	cpi	col, LCDCOLS
	breq	loop
	ldi	temp, ' '
	rcall	lcd_char
	inc	col
	rjmp	loop

;--------------------

char_data:
	;printable char: save to string buffer
	mov	YL, bufptr
	st	Y, temp
	inc	bufptr
	cpi	bufptr, Buffer+LCDCOLS
	brlo	loop
	
	;buffer full: received an entire line without space
	ldi	count, LCDCOLS
	rcall	display_string
	
	rjmp	loop
	
;===============================================================================

getc:
	;wait for char from UART
	sbis	UCSRA, RXC
	rjmp	getc
	in	temp, UDR
	ret
	
;--------------------

display_string:
	sbi	LCD_Ctrl, LCD_BACKLIGHT
	mov	temp, col ;check if string fits into current line
	add	temp, count
	cpi	temp, LCDCOLS+1
	brlo	ram2lcd
	cpi	line, LCDLINES-1 ;last line: use completely
	breq	ram2lcd
	inc	line ;use next line
	ldi	col, 0
	rcall	lcd_line
	
ram2lcd:
	;copy string buffer to LCD
;	ldi	count, LCDCOLS ;find # of chars to copy
;	sub	count, col
;	mov	temp, bufptr
;	subi	temp, Buffer
;	cp	temp, count
;	brsh	PC+2
;	mov	count, temp
;	cpi	count, 0 ;abort if copy length is 0
;	breq	ram2lcd_end
	ldi	YL, Buffer
ram2lcd_loop:
	cpi	bufptr, Buffer
	breq	ram2lcd_end
	cpi	col, LCDCOLS ;abort if line is full
	brsh	ram2lcd_eol
	ld	temp, Y+
	rcall	lcd_char
	inc	col
;	dec	count
;	brne	ram2lcd_loop
	dec	bufptr
	rjmp	ram2lcd_loop
	
ram2lcd_end:
	ldi	bufptr, Buffer ;clear buffer
	ret
	
ram2lcd_eol:
	;end of last line: display right arrow at lower right display corner
	ldi	col, LCDCOLS-1
	rcall	lcd_line
	ldi	temp, 0x00 ;custom char #0
	rcall	lcd_char
	ldi	line, LCDLINES ;lock display
	ret
	
;===============================================================================

lcd_line:
	;move cursor to column $col in line $line
	ldi	ZL, LOW(linetab*2)
	ldi	ZH, HIGH(linetab*2)
	add	ZL, line
	ldi	temp, 0
	adc	ZH, temp
	lpm	temp, Z
	ori	temp, 0x80
	add	temp, col
	rcall	lcd_cmd
	ret

;--------------------

lcd_init:
	;initialize LCD
	ldi	count, 20
lcd_init_wait:
	ldi	temp, 250 ;500ms (min. 15ms)
	rcall	lcdwait
	dec	count
	brne	lcd_init_wait
	ldi	temp, 250
	rcall	lcdwait
	ldi	temp, 0x03 ;function set #1
	rcall	lcd_cmd
	ldi	temp, 200 ;20ms (min. 4.1ms)
	rcall	lcdwait
	ldi	temp, 0x03 ;function set #2
	rcall	lcd_cmd
	ldi	temp, 100 ;10ms (min. 100us)
	rcall	lcdwait
	ldi	temp, 0x03 ;function set #3
	rcall	lcd_cmd
	ldi	temp, 100 ;10ms (min. 0)
	rcall	lcdwait
	ldi	temp, 0x38 ;function set: 8 bit, 2 lines, 5x7 font
	rcall	lcd_cmd
	ldi	temp, 100 ;10ms (min. 0)
	rcall	lcdwait
	ldi	temp, 0x0C ;display on, cursor off
	rcall	lcd_cmd
	ldi	temp, 0x06 ;increment cursor, no shifting
	rcall	lcd_cmd
	ldi	temp, 0x02 ;return home, return display from shifting
	rcall	lcd_cmd
	ldi	temp, 30 ;3ms (min. 1.52ms)
	rcall	lcdwait
	ldi	temp, 0x40 ;custom char #0: right arrow
	rcall	lcd_cmd
	ldi	temp, 0x00
	rcall	lcd_char
	ldi	temp, 0x04
	rcall	lcd_char
	ldi	temp, 0x02
	rcall	lcd_char
	ldi	temp, 0x1f
	rcall	lcd_char
	ldi	temp, 0x02
	rcall	lcd_char
	ldi	temp, 0x04
	rcall	lcd_char
	ldi	temp, 0x00
	rcall	lcd_char
	ldi	temp, 0x00
	rcall	lcd_char
	ldi	temp, 0x01 ;clear display
	rcall	lcd_cmd
	ldi	temp, 30 ;3ms (min. 1.52ms)
	rcall	lcdwait
	ret

;--------------------

lcd_clear:
	;clear entire LCD (both controllers)
	ldi	line, 0
	ldi	temp, 0x01 ;clear display
	rcall	lcd_cmd
	.if ( DualController )
	ldi	line, LCDLINES/2
	ldi	temp, 0x01 ;clear display
	rcall	lcd_cmd
	.endif
	ldi	temp, 30 ;3ms (min. 1.52ms)
	rcall	lcdwait
	ret

;--------------------

lcd_cmd:
	;send command to the LCD
	out	LCD_Data, temp
	cbi	LCD_Ctrl, LCD_RS
	rjmp	PC+1 ;1us
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	.if ( DualController )
	cpi	line, LCDLINES/2
	brsh	PC+2
	.endif
	sbi	LCD_Ctrl, LCD_E
	.if ( DualController )
	brlo	PC+2
	sbi	LCD_Ctrl, LCD_E2
	.endif
	rjmp	PC+1 ;1us
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	cbi	LCD_Ctrl, LCD_E
	cbi	LCD_Ctrl, LCD_E2
	ldi	temp, 2 ;100us (min. 37us)
	rcall	lcdwait
	ret

;--------------------

lcd_char:
	;send data byte to the LCD
	out	LCD_Data, temp
	sbi	LCD_Ctrl, LCD_RS
	rjmp	PC+1 ;1us
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	.if ( DualController )
	cpi	line, LCDLINES/2
	brsh	PC+2
	.endif
	sbi	LCD_Ctrl, LCD_E
	.if ( DualController )
	brlo	PC+2
	sbi	LCD_Ctrl, LCD_E2
	.endif
	rjmp	PC+1 ;1us
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	cbi	LCD_Ctrl, LCD_E
	cbi	LCD_Ctrl, LCD_E2
	ldi	temp, 1 ;100us (min. 37us)
	rcall	lcdwait
	ret

;--------------------

lcdwait:
	;wait temp * ~100us
	clr	wait
lcdwait_loop:
	dec	wait
	brne	lcdwait_loop
	dec	temp
	brne	lcdwait_loop
	ret
	